//
// Author:    Jae Chung
// Date:      10/20/99
// File Name: cbt.cc
//

#ifndef lint
static const char rcsid[] =
    "@(#) $Header: ns-2/cbt-dynamic.cc,v 1.0 1999/10/20 01:24:29 gnguyen Exp $ (LBL)";
#endif

#include <math.h>
#include <string.h>
#include <sys/types.h>
#include "template.h"
#include "random.h"
#include "flags.h"
#include "delay.h"
#include "cbt-dynamic.h"

static class DynamicCBTClass : public TclClass {
public:
	DynamicCBTClass() : TclClass("Queue/CBT/Dynamic") {}
	TclObject* create(int, const char*const*) {
		return (new DynamicCBTQueue);
	}
} class_dynamic_cbt;


DynamicCBTQueue::DynamicCBTQueue()
{
  bind_bool("chips_", &chips_);
  vq_ = new PacketQueue();
}


// Return the next packet in the queue for transmission.
Packet* DynamicCBTQueue::deque()
{
        ////////////////////////////////////
        // Normal dequeue (RED) procedure //
        ////////////////////////////////////
        Packet *p;
        p = q_->deque();

        if (p != 0) {
                idle_ = 0;
                bcount_ -= ((hdr_cmn*)p->access(off_cmn_))->size();
		if(chips_) {
		      Packet *vpkt = vq_->deque();
		      cbt_accnt_dequeue(vpkt);
		      free(vpkt);
		}
                else {
		  cbt_accnt_dequeue(p); // CBT accnt dequeue
		}
        } else {
                idle_ = 1;
                // deque() may invoked by Queue::reset at init
                // time (before the scheduler is instantiated).
                // deal with this case
                if (&Scheduler::instance() != NULL)
                        idletime_ = Scheduler::instance().clock();
                else
                        idletime_ = 0.0;
        }


        ///////////////////////////////////////////////////
        // Recording start of MM Logical Queue Idle time //
        ///////////////////////////////////////////////////
        if(mmu_class_.cq_size == 0) {
             if(mm_idle_ == 0) {
                  mm_idle_ = 1;
                  if(&Scheduler::instance() != NULL)
                        mm_idletime_ = Scheduler::instance().clock();
                  else
                        mm_idletime_ = 0.0;
             }
        }
        else {
             mm_idle_ = 0;
        }
    

        ////////////////////////////////////////////////////
        // Recording start of UDP Logical Queue Idle time //
        ////////////////////////////////////////////////////
	if(udp_class_.cq_size == 0) {
	     if(udp_idle_ == 0) {
	          udp_idle_ = 1;
	          if(&Scheduler::instance() != NULL)
		        udp_idletime_ = Scheduler::instance().clock();
		  else
		        udp_idletime_ = 0.0;
	     }
	}
	else {
	     udp_idle_ = 0;
	}
    

        return (p);
}



#define	DTYPE_NONE	0	/* ok, no drop */
#define	DTYPE_FORCED	1	/* a "forced" drop */
#define	DTYPE_UNFORCED	2	/* an "unforced" (random) drop */
#define CHIPS_LIMIT     0.6     /* if mmu's share is less than this value do chips */

void DynamicCBTQueue::enque(Packet* pkt)
{
        /////////////////////////////////////////
        // Counting Each flows (TCP, UDP, MMU) //
        /////////////////////////////////////////

        ItemType flow;
	hdr_cmn* cmnhead = (hdr_cmn*)pkt->access(off_cmn_);
	hdr_ip* iphead = (hdr_ip*)pkt->access(off_ip_);
	
	flow.des = iphead->dst();
	flow.fid = iphead->flowid();
	flow.time = Scheduler::instance().clock();

	if ((cmnhead->ptype() == PT_UDP) || (cmnhead->ptype() == PT_CBR)) {
	        // MM UDP packet
	        if(iphead->prio() == 15) {
		       mmu_counter.UpdateCount(flow);
		       //printf("%s  MMU: %f (%d)\n", name_, flow.time, mmu_counter.Count());
		} 
		// UDP packet
		else {
		      udp_counter.UpdateCount(flow);
		      //printf("%s  UDP: %f (%d)\n", name_, flow.time, udp_counter.Count());
		}
	}
	if ((cmnhead->ptype() == PT_TCP) || (cmnhead->ptype() == PT_ACK)) {
	        tcp_counter.UpdateCount(flow);
		//printf("%s  TCP: %f (%d)\n", name_, flow.time, tcp_counter.Count());
	}


	////////////////////////////////////////
	// Calculate the share of MMU and UDP //
	////////////////////////////////////////

	double mmu_share = 1/3;
	double udp_share = 1/3;
	int tot_num_flow = mmu_counter.Count() + udp_counter.Count() + tcp_counter.Count();
	if (tot_num_flow != 0) {
	    mmu_share = (double)(mmu_counter.Count()) / tot_num_flow;
	    udp_share = (double)(udp_counter.Count()) / tot_num_flow;
	}
	//printf("mmu_share = %f\n", mmu_share);
	//printf("udp_share = %f\n", udp_share);


	///////////////////////////////
	// Calculate Avg Queue Sizes //
	///////////////////////////////

	/*
	 * Calculating MM logical Queue's Average Length
	 */

        int mm_m = 0;
	if (mm_idle_) {
	        double mm_now = Scheduler::instance().clock();
		mm_idle_ = 0;
		
		if(link_) mm_ptc_ = (link_band_ * mmu_share) / (8. * edp_.mean_pktsize);
		// printf("mm_ptc_ = %f\n", mm_ptc_);
		mm_m = int(mm_ptc_ * (mm_now - mm_idletime_));
	} 
	calc_mm_q_avg(mm_m + 1);


	/*
	 * Calculating avg_q_len of UDP logical queue
	 */

        int udp_m = 0;
	if (udp_idle_) {
	        double udp_now = Scheduler::instance().clock();
		udp_idle_ = 0;
		udp_m = int(udp_ptc_ * (udp_now - udp_idletime_));
	} 
	calc_udp_q_avg(udp_m + 1);


	/*
	 * Calculating RED average
	 */

	int m = 0;
	if (idle_) {
		double now = Scheduler::instance().clock();
		/* To account for the period when the queue was empty. */
		idle_ = 0;
		m = int(edp_.ptc * (now - idletime_));
	}
	run_estimator(qib_ ? bcount_ : q_->length(), m + 1);

	printf("%s\t%f\t%d\t%f\t%f\t%f\n", name_, Scheduler::instance().clock(), 
	       (int)q_->length(), (double)edv_.v_ave, mm_q_avg_, udp_q_avg_);



	register double qavg = edv_.v_ave;
	int droptype = DTYPE_NONE;
	int qlen = qib_ ? bcount_ : q_->length();
	int qlim = qib_ ? (qlim_ * edp_.mean_pktsize) : qlim_;
	int red_congestion = 0;

	curq_ = qlen;	// helps to trace queue during arrival, if enabled


	/////////////
	//CBT Test //
	/////////////

	/*
	 * Drop non-TCP packets if class avg queue size over threshold
	 */

	hdr_cmn* cmnh = (hdr_cmn*)pkt->access(off_cmn_);
	hdr_ip* iph = (hdr_ip*)pkt->access(off_ip_);

	if ((qavg >= edp_.th_min) && (qlen > 1) && (edv_.old != 0)) red_congestion = 1;
	if (red_congestion && ((cmnh->ptype() == PT_UDP) || (cmnh->ptype() == PT_CBR))) {
	        // MM UDP packet
	        if(iph->prio() == 15) {
		        //double mmu_th = qavg * mmu_share * 1.05;
		        double mmu_th = qavg * mmu_share;
		        if(mm_q_avg_ >  mmu_th) {
			       drop(pkt);
			       return;
			}
	        }
		// UDP packet
		else {
		        double allowance = (edp_.th_max - edp_.th_min) / 10;
		        double udp_th = (edp_.th_min + allowance) * udp_share;
		        if(udp_q_avg_ > udp_th) {
			       drop(pkt);
			       return;
			}
		}
	}



	/////////////////////////////////////////////////////
	// RED Test (for non-TCP that pass CBT test & TCP) //
	/////////////////////////////////////////////////////

	/*
	 * count and count_bytes keeps a tally of arriving traffic
	 * that has not been dropped (i.e. how long, in terms of traffic,
	 * it has been since the last early drop)
	 */

	hdr_cmn* ch = (hdr_cmn*)pkt->access(off_cmn_);
	++edv_.count;
	edv_.count_bytes += ch->size();


	/*
	 * Mark Packet for Drop
	 */

	if (qavg >= edp_.th_min && qlen > 1) {
	       if (qavg >= edp_.th_max) {
		      droptype = DTYPE_FORCED;
	       } else if (edv_.old == 0) {
		      // SALLY: would like a comment here
		      edv_.count = 1;
		      edv_.count_bytes = ch->size();
		      edv_.old = 1;
	       } else if (drop_early(pkt))
		      droptype = DTYPE_UNFORCED;
	} else {
	       edv_.v_prob = 0.0;
	       edv_.old = 0;         // explain
	}

	/*
	 * Physical Queue Overflow Test
	 */
	if (qlen >= qlim) {
		// see if we've exceeded the queue size
	        droptype = DTYPE_FORCED;
	}


	/*
	 * Drop marked packets
	 */

	if (droptype == DTYPE_UNFORCED) {
	        if (de_drop_ != NULL)
		        de_drop_->recv(pkt);
		else
			drop(pkt);
	} else {
	        // enqueue pkt
	        if (chips_) {
		       Packet* vpkt = pkt->copy();
		       vq_->enque(vpkt);
		       
		       // if MM UDP packet and mmu_share < CHIPS_LIMIT do insert
		       if (((cmnh->ptype() == PT_UDP) || (cmnh->ptype() == PT_CBR)) 
			   && (iph->prio() == 15) && (mmu_share < CHIPS_LIMIT)) {
			      if ((int)qavg < (int)edp_.th_min)
				    q_->insert(pkt, (int)edp_.th_min);
			      else
				    q_->insert(pkt, (int)qavg);
		       }
		       else {
			      q_->enque(pkt);
		       }
		}
		else {
		       q_->enque(pkt);
		}
		bcount_ += ch->size();
		cbt_accnt_enqueue(pkt); // CBT accnt enqueue

		// drop a packet if we were told to
                if (droptype == DTYPE_FORCED) {
                       /* drop random victim or last one */
                       int victim_off = PickPacketToDrop();

		       pkt = q_->lookup(victim_off);
                       q_->remove(pkt);
                       bcount_ -= ((hdr_cmn*)pkt->access(off_cmn_))->size();

		       if(chips_) {
			      Packet *vpkt = vq_->lookup(victim_off);
			      vq_->remove(vpkt);
			      cbt_accnt_dequeue(vpkt);
			      free(vpkt);
		       }
		       else {
			      cbt_accnt_dequeue(pkt); // CBT accnt dequeue
		       }

		       drop(pkt);
                }
	}
	return;
}


int DynamicCBTQueue::PickPacketToDrop() 
{
        int victim;

        if (drop_front_)
                victim = min(1, q_->length()-1);
        else if (drop_rand_)
                victim = Random::integer(q_->length());
        else                    /* default is drop_tail_ */
                victim = q_->length() - 1;

        return(victim); 
}
