//
// Author:    Jae Chung
// Date:      10/20/99
// File Name: cbt.cc
//

#ifndef lint
static const char rcsid[] =
    "@(#) $Header: ns-2/cbt.cc,v 1.0 1999/10/20 01:24:29 gnguyen Exp $ (LBL)";
#endif

#include <math.h>
#include <string.h>
#include <sys/types.h>
#include "template.h"
#include "random.h"
#include "flags.h"
#include "delay.h"
#include "cbt.h"

static class CBTClass : public TclClass {
public:
	CBTClass() : TclClass("Queue/CBT") {}
	TclObject* create(int, const char*const*) {
		return (new CBTQueue);
	}
} class_cbt;

CBTQueue::CBTQueue() : link_(NULL), bcount_(0), de_drop_(NULL),
	tchan_(0), idle_(1)
{
	bind_bool("bytes_", &edp_.bytes);	    // boolean: use bytes?
	bind_bool("queue-in-bytes_", &qib_);	    // boolean: q in bytes?
	bind("thresh_", &edp_.th_min);		    // minthresh
	bind("maxthresh_", &edp_.th_max);	    // maxthresh
	bind("mean_pktsize_", &edp_.mean_pktsize);  // avg pkt size
	bind("q_weight_", &edp_.q_w);		    // for EWMA
	bind_bool("wait_", &edp_.wait);
	bind("linterm_", &edp_.max_p_inv);
	bind_bool("setbit_", &edp_.setbit);	    // mark instead of drop
	bind_bool("drop-tail_", &drop_tail_);	    // drop last pkt
	bind_bool("drop-front_", &drop_front_);	    // drop first pkt
	bind_bool("drop-rand_", &drop_rand_);	    // drop pkt at random

	bind("ave_", &edv_.v_ave);		    // average queue sie
	bind("prob1_", &edv_.v_prob1);		    // dropping probability
	bind("curq_", &curq_);			    // current queue size


	bind("th_udp_", &udp_class_.th_max);        // CBT: udp_class threshold
	bind("th_mmu_", &mmu_class_.th_max);        // CBT: mmu_class threshold
	bind("bypass_red_", &bypass_red_);          // CBT: Bypass RED? (default: 0)

	bind("mm_q_avg_", &mm_q_avg_);
	bind("mm_q_weight_", &mm_q_weight_);
	bind("udp_q_avg_", &udp_q_avg_);
	bind("udp_q_weight_", &udp_q_weight_);


	bind("off_ip_", &off_ip_);                  // IP header offset

	q_ = new PacketQueue();			    // underlying queue
	pq_ = q_;
	reset();

	// print_edp();
	// print_edv();
	// print_cbt();
}


void CBTQueue::reset()
{
	// If queue is measured in bytes, scale min/max thresh
	// by the size of an average packet (which is specified by user)
	if (qib_) {
		edp_.th_min *= edp_.mean_pktsize;
		edp_.th_max *= edp_.mean_pktsize;
	}

	// Compute the "packet time constant" if we know the
	// link bandwidth.  The ptc is the max number of (avg sized)
	// pkts per second which can be placed on the link.
	// The link bw is given in bits/sec, so scale mean psize
	// accordingly.
	//
	// Reset CBT: MM mean packet throughput (pkt/sec) 
	// Assume 1/3 of total bandwidth is dedicated to MM

	link_band_ = 0;
	edp_.ptc = 0;
	mm_ptc_ = 0;
	udp_ptc_ = 0;
	if (link_) {
	        link_band_ = link_->bandwidth();
		edp_.ptc = link_band_ / (8. * edp_.mean_pktsize);
		mm_ptc_  = (link_band_ / 3) / (8. * edp_.mean_pktsize);
		udp_ptc_ = (link_band_ / 3) / (8. * edp_.mean_pktsize);
	}

	edv_.v_ave = 0.0;
	edv_.v_slope = 0.0;
	edv_.count = 0;
	edv_.count_bytes = 0;
	edv_.old = 0;
	edv_.v_a = 1 / (edp_.th_max - edp_.th_min);
	edv_.v_b = - edp_.th_min / (edp_.th_max - edp_.th_min);


	// Reset RED: RED Queue's Idle status
	// Reset CBT: MM Queue's Idle status
	idle_ = 1;
	mm_idle_ = 1;
	if (&Scheduler::instance() != NULL) {
	       idletime_ = Scheduler::instance().clock();
	       mm_idletime_ = Scheduler::instance().clock();
	       udp_idletime_ = Scheduler::instance().clock();
	}
	else {
	       idletime_ = 0.0; // sched not instantiated yet
	       mm_idletime_ = 0.0; // sched not instantiated yet
	       udp_idletime_ = 0.0; // sched not instantiated yet
	}

	Queue::reset();

	bcount_ = 0;
	udp_class_.cq_size = 0;
	mmu_class_.cq_size = 0;

	if (link_) printf("link_bandwidth = %f\n", link_->bandwidth());
	printf("edp.mean_pktsize = %d\n", edp_.mean_pktsize);
	printf("edp_.ptc = %f   mm_ptc_ = %f   udp_ptc_ = %f\n", edp_.ptc, mm_ptc_, udp_ptc_);
}


// Compute the average queue size for RED (EWMA method)
// nqueued can be bytes or packets
void CBTQueue::run_estimator(int nqueued, int m)
{
        float f, f_sl, f_old;

        f = edv_.v_ave;
        f_sl = edv_.v_slope;
#define RED_EWMA
#ifdef RED_EWMA
        while (--m >= 1) {
                f_old = f;
                f *= 1.0 - edp_.q_w;
        }
        f_old = f;
        f *= 1.0 - edp_.q_w;
        f += edp_.q_w * nqueued;
#endif
#ifdef RED_HOLT_WINTERS
        while (--m >= 1) {
                f_old = f;
                f += f_sl;
                f *= 1.0 - edp_.q_w;
                f_sl *= 1.0 - 0.5 * edp_.q_w;
                f_sl += 0.5 * edp_.q_w * (f - f_old);
        }
        f_old = f;
        f += f_sl;
        f *= 1.0 - edp_.q_w;
        f += edp_.q_w * nqueued;
        f_sl *= 1.0 - 0.5 * edp_.q_w;
        f_sl += 0.5 * edp_.q_w * (f - f_old);
#endif
        edv_.v_ave = f;
        edv_.v_slope = f_sl;
}



void CBTQueue::calc_mm_q_avg(int m)
{
        double f = mm_q_avg_;

        while (--m >= 1) {
                f *= 1.0 - mm_q_weight_;
        }
        f *= 1.0 - mm_q_weight_;
        f += mm_q_weight_ * mmu_class_.cq_size;

	mm_q_avg_ = f;
}

void CBTQueue::calc_udp_q_avg(int m)
{
        double f = udp_q_avg_;

        while (--m >= 1) {
                f *= 1.0 - udp_q_weight_;
        }
        f *= 1.0 - udp_q_weight_;
        f += udp_q_weight_ * udp_class_.cq_size;

	udp_q_avg_ = f;
}



// Return the next packet in the queue for transmission.
Packet* CBTQueue::deque()
{
	////////////////////////////////////
	// Normal dequeue (RED) procedure //
	////////////////////////////////////
	Packet *p;
	p = q_->deque();

	if (p != 0) {
	        idle_ = 0;
		bcount_ -= ((hdr_cmn*)p->access(off_cmn_))->size();
		cbt_accnt_dequeue(p); // CBT accnt dequeue

	} else {
		idle_ = 1;
		// deque() may invoked by Queue::reset at init
		// time (before the scheduler is instantiated).
		// deal with this case
		if (&Scheduler::instance() != NULL)
			idletime_ = Scheduler::instance().clock();
		else
			idletime_ = 0.0;
	}


        ///////////////////////////////////////////////////
        // Recording start of MM Logical Queue Idle time //
        ///////////////////////////////////////////////////
	if(mmu_class_.cq_size == 0) {
	     if(mm_idle_ == 0) {
	          mm_idle_ = 1;
	          if(&Scheduler::instance() != NULL)
		        mm_idletime_ = Scheduler::instance().clock();
		  else
		        mm_idletime_ = 0.0;
	     }
	}
	else {
	     mm_idle_ = 0;
	}
    

        ////////////////////////////////////////////////////
        // Recording start of UDP Logical Queue Idle time //
        ////////////////////////////////////////////////////
	if(udp_class_.cq_size == 0) {
	     if(udp_idle_ == 0) {
	          udp_idle_ = 1;
	          if(&Scheduler::instance() != NULL)
		        udp_idletime_ = Scheduler::instance().clock();
		  else
		        udp_idletime_ = 0.0;
	     }
	}
	else {
	     udp_idle_ = 0;
	}
    
	return (p);
}


// should the packet be dropped/marked due to a probabilistic drop?
int CBTQueue::drop_early(Packet* pkt)
{
	hdr_cmn* ch = (hdr_cmn*)pkt->access(off_cmn_);

	double p = edv_.v_a * edv_.v_ave + edv_.v_b;
	p /= edp_.max_p_inv;
	edv_.v_prob1 = p;
	if (edv_.v_prob1 > 1.0)
		edv_.v_prob1 = 1.0;
	double count1 = edv_.count;
	if (edp_.bytes)
		count1 = (double) (edv_.count_bytes/edp_.mean_pktsize);
	if (edp_.wait) {
		if (count1 * p < 1.0)
			p = 0.0;
		else if (count1 * p < 2.0)
			p /= (2 - count1 * p);
		else
			p = 1.0;
	} else {
		if (count1 * p < 1.0)
			p /= (1.0 - count1 * p);
		else
			p = 1.0;
	}
	if (edp_.bytes && p < 1.0) {
		p = p * ch->size() / edp_.mean_pktsize;
	}
	if (p > 1.0)
		p = 1.0;
	edv_.v_prob = p;

	// drop probability is computed, pick random number and act
	double u = Random::uniform();
	if (u <= edv_.v_prob) {
		// DROP or MARK
		edv_.count = 0;
		edv_.count_bytes = 0;
		hdr_flags* hf = (hdr_flags*)pickPacketForECN(pkt)->access(off_flags_);
		if (edp_.setbit && hf->ect()) {
			hf->ce() = 1; 	// mark Congestion Experienced bit
			return (0);	// no drop
		} else {
			return (1);	// drop
		}
	}
	return (0);			// no DROP/mark
}

/*
 * Pick packet for early congestion notification (ECN). This packet is then
 * marked or dropped. Having a separate function do this is convenient for
 * supporting derived classes that use the standard RED algorithm to compute
 * average queue size but use a different algorithm for choosing the packet for 
 * ECN notification.
 */
Packet* CBTQueue::pickPacketForECN(Packet* pkt)
{
	return pkt; /* pick the packet that just arrived */
}

/*
 * Pick packet to drop. Having a separate function do this is convenient for
 * supporting derived classes that use the standard RED algorithm to compute
 * average queue size but use a different algorithm for choosing the victim.
 */
Packet* CBTQueue::pickPacketToDrop() 
{
	int victim;

	if (drop_front_)
		victim = min(1, q_->length()-1);
	else if (drop_rand_)
		victim = Random::integer(q_->length());
	else			/* default is drop_tail_ */
		victim = q_->length() - 1;

	return(q_->lookup(victim)); 
}

/*
 * Receive a new packet arriving at the queue.
 * The average queue size is computed.  If the average size
 * exceeds the threshold, then the dropping probability is computed,
 * and the newly-arriving packet is dropped with that probability.
 * The packet is also dropped if the maximum queue size is exceeded.
 *
 * "Forced" drops mean a packet arrived when the underlying queue was
 * full or when the average q size exceeded maxthresh.
 * "Unforced" means a RED random drop.
 *
 * For forced drops, either the arriving packet is dropped or one in the
 * queue is dropped, depending on the setting of drop_tail_.
 * For unforced drops, the arriving packet is always the victim.
 */

#define	DTYPE_NONE	0	/* ok, no drop */
#define	DTYPE_FORCED	1	/* a "forced" drop */
#define	DTYPE_UNFORCED	2	/* an "unforced" (random) drop */

void CBTQueue::enque(Packet* pkt)
{
	/////////////
	//CBT Test //
	/////////////

	/*
	 * Drop non-TCP packets if class avg queue size over threshold
	 */

	hdr_cmn* cmnh = (hdr_cmn*)pkt->access(off_cmn_);
	hdr_ip* iph = (hdr_ip*)pkt->access(off_ip_);

	if ((cmnh->ptype() == PT_UDP) || (cmnh->ptype() == PT_CBR)) {
	        // MM UDP packet
	        if(iph->prio() == 15) { 

		        /*
		         * Calculating avg_q_len of MM logical queue
		         */
		        int mm_m = 0;
			if (mm_idle_) {
			        double mm_now = Scheduler::instance().clock();
			        mm_idle_ = 0;
			        mm_m = int(mm_ptc_ * (mm_now - mm_idletime_));
			} 
			calc_mm_q_avg(mm_m + 1);

			/*
			 * MM Threshold Test
			 */
		        if(mm_q_avg_ > (double) mmu_class_.th_max) {
			       printf("%s\t%f\t%d\t%f\t%f\t%f\n", name_, 
				      Scheduler::instance().clock(), 
				      (int)q_->length(), (double)edv_.v_ave, 
				      mm_q_avg_, udp_q_avg_);
			       drop(pkt);
			       return;
			}
	        }
		// UDP packet
		else {

		       /*
		        * Calculating avg_q_len of UDP logical queue
		        */
		       int udp_m = 0;
		       if (udp_idle_) {
		               double udp_now = Scheduler::instance().clock();
		               udp_idle_ = 0;
		               udp_m = int(udp_ptc_ * (udp_now - udp_idletime_));
		       } 
		       calc_udp_q_avg(udp_m + 1);

		       /*
			* UDP Threshold Test
			*/
		       if(udp_q_avg_ >= (double) udp_class_.th_max) {
			       printf("%s\t%f\t%d\t%f\t%f\t%f\n", name_, 
				      Scheduler::instance().clock(), 
				      (int)q_->length(), (double)edv_.v_ave, 
				      mm_q_avg_, udp_q_avg_);
			       drop(pkt);
		               return;
		       }
		}
	}

	
	/////////////////////////////////////////////////////
	// RED Test (for non-TCP that pass CBT test & TCP) //
	/////////////////////////////////////////////////////

	/*
	 * Calculating RED average
	 */

	int m = 0;
	if (idle_) {
		double now = Scheduler::instance().clock();
		/* To account for the period when the queue was empty. */
		idle_ = 0;
		m = int(edp_.ptc * (now - idletime_));
	}
	run_estimator(qib_ ? bcount_ : q_->length(), m + 1);

	printf("%s\t%f\t%d\t%f\t%f\t%f\n", name_, Scheduler::instance().clock(), 
	       (int)q_->length(), (double)edv_.v_ave, mm_q_avg_, udp_q_avg_);


	/*
	 * count and count_bytes keeps a tally of arriving traffic
	 * that has not been dropped (i.e. how long, in terms of traffic,
	 * it has been since the last early drop)
	 */

	hdr_cmn* ch = (hdr_cmn*)pkt->access(off_cmn_);
	++edv_.count;
	edv_.count_bytes += ch->size();


	register double qavg = edv_.v_ave;
	int droptype = DTYPE_NONE;
	int qlen = qib_ ? bcount_ : q_->length();
	int qlim = qib_ ? (qlim_ * edp_.mean_pktsize) : qlim_;

	curq_ = qlen;	// helps to trace queue during arrival, if enabled


	/*
	 * Mark Packet for Drop
	 */

	if (qavg >= edp_.th_min && qlen > 1) {
	       if (qavg >= edp_.th_max) {
		      droptype = DTYPE_FORCED;
	       } else if (edv_.old == 0) {
		      // SALLY: would like a comment here
		      edv_.count = 1;
		      edv_.count_bytes = ch->size();
		      edv_.old = 1;
	       } else if (drop_early(pkt))
		      droptype = DTYPE_UNFORCED;
	} else {
	       edv_.v_prob = 0.0;
	       edv_.old = 0;		// explain
	}

	/*
	 * Physical Queue Overflow Test
	 */

	if (qlen >= qlim)  {
	       // see if we've exceeded the queue size
	       droptype = DTYPE_FORCED;
	}

	/*
	 * Drop marked packets
	 */

	if (droptype == DTYPE_UNFORCED) {
		/* pick packet for ECN, which is dropping in this case */
		Packet *pkt_to_drop = pickPacketForECN(pkt);
		/* 
		 * If the packet picked is different that the one that just arrived,
		 * add it to the queue and remove the chosen packet.
		 */
		if (pkt_to_drop != pkt) {
			q_->enque(pkt);
			bcount_ += ch->size();
			cbt_accnt_enqueue(pkt); // CBT accnt enqueue
			
			q_->remove(pkt_to_drop);
			bcount_ -= ((hdr_cmn*)pkt_to_drop->access(off_cmn_))->size();
			cbt_accnt_dequeue(pkt_to_drop); //CBT accnt dequeue

			pkt = pkt_to_drop; /* XXX okay because pkt is not needed anymore */
		}
		// deliver to special "edrop" target, if defined
		if (de_drop_ != NULL)
			de_drop_->recv(pkt);
		else
			drop(pkt);
	} else {
		/* forced drop, or not a drop: first enqueue pkt */
		q_->enque(pkt);
		bcount_ += ch->size();
		cbt_accnt_enqueue(pkt); // CBT accnt enqueue

		/* drop a packet if we were told to */
		if (droptype == DTYPE_FORCED) {
			/* drop random victim or last one */
			pkt = pickPacketToDrop();
			q_->remove(pkt);
			bcount_ -= ((hdr_cmn*)pkt->access(off_cmn_))->size();
			cbt_accnt_dequeue(pkt); // CBT accnt dequeue

			drop(pkt);
		}
	}
	return;
}



/*
 * CBT Enqueue: CBT logical queue accounting
 */
void CBTQueue::cbt_accnt_enqueue(Packet* pkt)
{
  hdr_cmn* cmnh = (hdr_cmn*)pkt->access(off_cmn_);
  hdr_ip* iph = (hdr_ip*)pkt->access(off_ip_);

  if((cmnh->ptype() == PT_UDP) || (cmnh->ptype() == PT_CBR)) {
        // MM UDP packet
        if(iph->prio() == 15) { 
	      mmu_class_.cq_size++;
	      mmu_class_.cq_size_byte += cmnh->size();
              // printf("Enqueue: MMUDP (%d)\n", mmu_class_.cq_size);
	}
	// UDP packet
	else {
	      udp_class_.cq_size++;
	      udp_class_.cq_size_byte += cmnh->size();
	      // printf("Enqueue: UDP (%d)\n", udp_class_.cq_size);
	}
  }
}


/*
 * CBT Dequeue: CBT logical queue accounting
 */
void CBTQueue::cbt_accnt_dequeue(Packet* pkt)
{
  hdr_cmn* cmnh = (hdr_cmn*)pkt->access(off_cmn_);
  hdr_ip* iph = (hdr_ip*)pkt->access(off_ip_);

  if((cmnh->ptype() == PT_UDP) || (cmnh->ptype() == PT_CBR)) {
        // MM UDP packet
        if(iph->prio() == 15) { 
	      mmu_class_.cq_size--;
	      mmu_class_.cq_size_byte -= cmnh->size();
              // printf("Dequeue: MMUDP (%d)\n", mmu_class_.cq_size);
	}
	// Other UDP packet
	else {
	      udp_class_.cq_size--;
	      udp_class_.cq_size_byte -= cmnh->size();
	      // printf("Dequeue: UDP (%d)\n", udp_class_.cq_size);
	}
  }
}



int CBTQueue::command(int argc, const char*const* argv)
{
	Tcl& tcl = Tcl::instance();
	if (argc == 2) {
		if (strcmp(argv[1], "reset") == 0) {
			reset();
			return (TCL_OK);
		}
		if (strcmp(argv[1], "early-drop-target") == 0) {
			if (de_drop_ != NULL)
				tcl.resultf("%s", de_drop_->name());
			return (TCL_OK);
		}
	} else if (argc == 3) {
		// attach a file for variable tracing
		if (strcmp(argv[1], "attach") == 0) {
			int mode;
			const char* id = argv[2];
			tchan_ = Tcl_GetChannel(tcl.interp(), (char*)id, &mode);
			if (tchan_ == 0) {
				tcl.resultf("CBT: trace: can't attach %s for writing", id);
				return (TCL_ERROR);
			}
			return (TCL_OK);
		}
		// tell CBT about link stats
		if (strcmp(argv[1], "link") == 0) {
			LinkDelay* del = (LinkDelay*)TclObject::lookup(argv[2]);
			if (del == 0) {
				tcl.resultf("CBT: no LinkDelay object %s",
					argv[2]);
				return(TCL_ERROR);
			}

			// set ptc now
			link_ = del;
			link_band_ = link_->bandwidth();
			edp_.ptc = link_band_ / (8. * edp_.mean_pktsize);
			mm_ptc_  = (link_band_ / 3) / (8. * edp_.mean_pktsize);
			udp_ptc_ = (link_band_ / 3) / (8. * edp_.mean_pktsize);
			return (TCL_OK);
		}
		if (strcmp(argv[1], "early-drop-target") == 0) {
			NsObject* p = (NsObject*)TclObject::lookup(argv[2]);
			if (p == 0) {
				tcl.resultf("no object %s", argv[2]);
				return (TCL_ERROR);
			}
			de_drop_ = p;
			return (TCL_OK);
		}
		if (!strcmp(argv[1], "packetqueue-attach")) {
			delete q_;
			if (!(q_ = (PacketQueue*) TclObject::lookup(argv[2])))
				return (TCL_ERROR);
			else {
				pq_ = q_;
				return (TCL_OK);
			}
		}
	}
	return (Queue::command(argc, argv));
}


/*
 * Routine called by TracedVar facility when variables change values.
 * Currently used to trace values of avg queue size, drop probability,
 * and the instantaneous queue size seen by arriving packets.
 * Note that the tracing of each var must be enabled in tcl to work.
 */

void CBTQueue::trace(TracedVar* v)
{
	char wrk[500], *p;

	if (((p = strstr(v->name(), "ave")) == NULL) &&
	    ((p = strstr(v->name(), "prob")) == NULL) &&
	    ((p = strstr(v->name(), "curq")) == NULL)) {
		fprintf(stderr, "CBT:unknown trace var %s\n",
			v->name());
		return;
	}

	if (tchan_) {
		int n;
		double t = Scheduler::instance().clock();
		// XXX: be compatible with nsv1 RED trace entries
		if (*p == 'c') {
			sprintf(wrk, "Q %g %d", t, int(*((TracedInt*) v)));
		} else {
			sprintf(wrk, "%c %g %g", *p, t,
				double(*((TracedDouble*) v)));
		}
		n = strlen(wrk);
		wrk[n] = '\n'; 
		wrk[n+1] = 0;
		(void)Tcl_Write(tchan_, wrk, n+1);
	}
	return; 
}

/* for debugging help */
void CBTQueue::print_edp()
{
	printf("mean_pktsz: %d\n", edp_.mean_pktsize); 
	printf("bytes: %d, wait: %d, setbit: %d\n",
	       edp_.bytes, edp_.wait, edp_.setbit);
	printf("minth: %f, maxth: %f\n", edp_.th_min, edp_.th_max);
	printf("max_p_inv: %f, qw: %f, ptc: %f\n",
	       edp_.max_p_inv, edp_.q_w, edp_.ptc);
	printf("qlim: %d, idletime: %f\n", qlim_, idletime_);
	printf("=========\n");
}

void CBTQueue::print_edv()
{
	printf("v_a: %f, v_b: %f\n", edv_.v_a, edv_.v_b);
}

void CBTQueue::print_cbt()
{
        printf("<CBT>  bypass_red_ = %d\n", bypass_red_);
        printf("    UDP:      th_max = %f,  cq_size = %d,  cq_size_byte = %d\n", 
	       udp_class_.th_max, udp_class_.cq_size, udp_class_.cq_size_byte);
	printf("    MM-UDP:   th_max = %f,  cq_size = %d,  cq_size_byte = %d\n", 
	       mmu_class_.th_max, mmu_class_.cq_size, mmu_class_.cq_size_byte);
}
